package IdPAccountManager::Logger;

use strict;
use warnings;

use base qw(Exporter);

use English qw(-no_match_vars);
use POSIX qw(strftime);

use constant {
    LOG_DEBUG  => 0,
    LOG_INFO   => 1,
    LOG_TRACE  => 1,
    LOG_NOTICE => 2,
    LOG_ERROR  => 3,
};

our @EXPORT = qw(LOG_DEBUG LOG_INFO LOG_TRACE LOG_NOTICE LOG_ERROR);

my @value2name = qw(DEBUG INFO NOTICE ERROR);

my %name2value = (
    debug  => LOG_DEBUG,
    info   => LOG_INFO,
    trace  => LOG_TRACE,
    notice => LOG_NOTICE,
    error  => LOG_ERROR,
);

sub new {
    my ($pkg, %args) = @_;

    die "missing argument 'file'"  unless defined $args{file};
    die "missing argument 'verbosity'" unless defined $args{verbosity};
    die "invalid argument 'verbosity'" unless exists $name2value{$args{verbosity}};

    my $handle;
    open($handle, '>>', $args{file}) or die "failed to open $args{file}: $ERRNO";

    my $self = {
        handle    => $handle,
        verbosity => $name2value{$args{verbosity}},
    };

    bless $self, $pkg;

    return $self;
}

sub log {
    my ($self, %args) = @_;

    die "missing argument 'message'" unless defined $args{message};
    die "missing argument 'level'"   unless defined $args{level};

    return if $args{level} < $self->{verbosity};

    ## Determine calling function and parameters
    ## If in 'err' level, build a stack trace
    my $caller_string;
    if ($args{level} == LOG_ERROR) {
        my $go_back = 1;
        my @calls;
        while (my @call = caller($go_back)) {
            unshift @calls, $call[3] . '#' . $call[2];
            $go_back++;
        }

        $caller_string = join(' > ', @calls);
    } else {
        my @call = caller(1);
        $caller_string = $call[3] . '()';
    }

    my $level   = $value2name[$args{level}];
    my $address = $ENV{REMOTE_ADDR} || 'NOIP';
    my $user    = $ENV{mail} || 'NOEMAIL';
    my $date    = POSIX::strftime("%Y:%m:%d %H:%M:%S", localtime(time));

    print
        {$self->{handle}}
        "$date - ($level) - $address - $user - $caller_string $args{message}\n";
}

sub DESTROY {
    my ($self) = @_;

    close $self->{handle};
}

1;
